#include <windows.h>
#include "../wdl/jnetlib/netinc.h"

#include "ping.h"

#define MAX_SIMUL_PINGS 32

int g_config_max_simul_pings=16;

typedef time_t n_time;
typedef short n_short;
typedef unsigned short u_int16_t;
typedef unsigned int u_int32_t;


#define ICMP_ECHO               8               /* echo service */
#define ICMP_ECHOREPLY          0               /* echo reply */
#define ICMP_MINLEN     8        

#pragma pack(push, 1)


struct icmp_ra_addr {
        u_int32_t ira_addr;
        u_int32_t ira_preference;
};

struct ip {
#ifdef _IP_VHL
        u_char  ip_vhl;                 /* version << 4 | header length >> 2 */
#else
        u_int   ip_hl:4,                /* header length */
                ip_v:4;                 /* version */
#endif /* not _IP_VHL */
        u_char  ip_tos;                 /* type of service */
        u_short ip_len;                 /* total length */
        u_short ip_id;                  /* identification */
        u_short ip_off;                 /* fragment offset field */
#define IP_RF 0x8000                    /* reserved fragment flag */
#define IP_DF 0x4000                    /* dont fragment flag */
#define IP_MF 0x2000                    /* more fragments flag */
#define IP_OFFMASK 0x1fff               /* mask for fragmenting bits */
        u_char  ip_ttl;                 /* time to live */
        u_char  ip_p;                   /* protocol */
        u_short ip_sum;                 /* checksum */
        struct  in_addr ip_src,ip_dst;  /* source and dest address */
};

struct icmp {
        u_char  icmp_type;              /* type of message, see below */
        u_char  icmp_code;              /* type sub code */
        u_short icmp_cksum;             /* ones complement cksum of struct */
        union {
                u_char ih_pptr;                 /* ICMP_PARAMPROB */
                struct in_addr ih_gwaddr;       /* ICMP_REDIRECT */
                struct ih_idseq {
                        n_short icd_id;
                        n_short icd_seq;
                } ih_idseq;
                int ih_void;

                /* ICMP_UNREACH_NEEDFRAG -- Path MTU Discovery (RFC1191) */
                struct ih_pmtu {
                        n_short ipm_void;
                        n_short ipm_nextmtu;
                } ih_pmtu;

                struct ih_rtradv {
                        u_char irt_num_addrs;
                        u_char irt_wpa;
                        u_int16_t irt_lifetime;
                } ih_rtradv;
        } icmp_hun;
#define icmp_pptr       icmp_hun.ih_pptr
#define icmp_gwaddr     icmp_hun.ih_gwaddr
#define icmp_id         icmp_hun.ih_idseq.icd_id
#define icmp_seq        icmp_hun.ih_idseq.icd_seq
#define icmp_void       icmp_hun.ih_void
#define icmp_pmvoid     icmp_hun.ih_pmtu.ipm_void
#define icmp_nextmtu    icmp_hun.ih_pmtu.ipm_nextmtu
#define icmp_num_addrs  icmp_hun.ih_rtradv.irt_num_addrs
#define icmp_wpa        icmp_hun.ih_rtradv.irt_wpa
#define icmp_lifetime   icmp_hun.ih_rtradv.irt_lifetime
        union {
                struct id_ts {
                        n_time its_otime;
                        n_time its_rtime;
                        n_time its_ttime;
                } id_ts;
                struct id_ip  {
                        struct ip idi_ip;
                        /* options and then 64 bits of data */
                } id_ip;
                struct icmp_ra_addr id_radv;
                u_int32_t id_mask;
                char    id_data[1];
        } icmp_dun;
#define icmp_otime      icmp_dun.id_ts.its_otime
#define icmp_rtime      icmp_dun.id_ts.its_rtime
#define icmp_ttime      icmp_dun.id_ts.its_ttime
#define icmp_ip         icmp_dun.id_ip.idi_ip
#define icmp_radv       icmp_dun.id_radv
#define icmp_mask       icmp_dun.id_mask
#define icmp_data       icmp_dun.id_data
};

#pragma pack(pop)

static int s;
static unsigned short ident;

typedef struct
{
  DWORD address;
  DWORD start_t;
  int timeout_t;
  int seq;
} pingRec;

static pingRec m_pings[MAX_SIMUL_PINGS];
static int m_pings_used;

void ping_init()
{
  struct protoent *proto;

	if ((proto = getprotobyname("icmp")) == NULL) return;
  s=socket(AF_INET,SOCK_RAW,proto->p_proto);

  if (!s) return;

  ident=(unsigned short) (GetCurrentProcessId()&0xffff);

  int on=1;
  setsockopt(s, SOL_SOCKET, SO_DONTROUTE, (char *)&on, sizeof(on));
  SET_SOCK_BLOCK(s,0);

  m_pings_used=0;

}

void ping_clear()
{
  m_pings_used=0;
}

int ping_avail()
{
  return min(MAX_SIMUL_PINGS,g_config_max_simul_pings)-m_pings_used;
}

static int ntransmitted;

#define	MAXPACKET	4096	/* max packet size */


static u_short in_cksum(u_short *addr, int len)
{
	register int nleft = len;
	register u_short *w = addr;
	register u_short answer;
	register int sum = 0;

	/*
	 *  Our algorithm is simple, using a 32 bit accumulator (sum),
	 *  we add sequential 16 bit words to it, and at the end, fold
	 *  back all the carry bits from the top 16 bits into the lower
	 *  16 bits.
	 */
	while( nleft > 1 )  {
		sum += *w++;
		nleft -= 2;
	}

	/* mop up an odd byte, if necessary */
	if( nleft == 1 ) {
		u_short	u = 0;

		*(u_char *)(&u) = *(u_char *)w ;
		sum += u;
	}

	/*
	 * add back carry outs from top 16 bits to low 16 bits
	 */
	sum = (sum >> 16) + (sum & 0xffff);	/* add hi 16 to low 16 */
	sum += (sum >> 16);			/* add carry */
	answer = ~sum;				/* truncate to 16 bits */
	return (answer);
}


void ping_add(DWORD address, int timeout_ms) // returns >= 0 on success (ms reply time) or <0 on failure
{
  if (ping_avail() < 1) return;
  if (!s) return;

  int datalen=64-8;

	char outpack[MAXPACKET];
	struct icmp *icp = (struct icmp *) outpack;
	int i, cc;
	char *datap = &outpack[8];

	icp->icmp_type = ICMP_ECHO;
	icp->icmp_code = 0;
	icp->icmp_cksum = 0;
	icp->icmp_seq = ++ntransmitted;
	icp->icmp_id = ident;		/* ID */

	cc = datalen+8;			/* skips ICMP portion */

	for( i=0; i<datalen; i++)	/* skip 8 for time */
		*datap++ = i;

	/* Compute ICMP checksum here */
	icp->icmp_cksum = in_cksum( (unsigned short *)icp, cc );

  struct sockaddr whereto;
	struct sockaddr_in *to = (struct sockaddr_in *) &whereto;

	memset((char *)&whereto, 0, sizeof(struct sockaddr) );
	to->sin_family = AF_INET;
	to->sin_addr.s_addr = (address);

  int pingpos=m_pings_used++;

  m_pings[pingpos].address = address;
  m_pings[pingpos].seq=icp->icmp_seq;
  m_pings[pingpos].timeout_t = timeout_ms;
  DWORD start_t=m_pings[pingpos].start_t=GetTickCount();

  for (;;)
  {
	  i = sendto( s, outpack, cc, 0, &whereto, sizeof(struct sockaddr) );

    if (i != WSAEWOULDBLOCK) break;
    if (GetTickCount()  > start_t+timeout_ms || GetTickCount() < start_t) 
    {
      m_pings[pingpos].start_t=0;
      m_pings[pingpos].timeout_t=1;
      return; // timed out while trying to send
    }
    Sleep(4);
  }
  if( i < 0 || i != cc) 
  {
    m_pings[pingpos].start_t=0;
    m_pings[pingpos].timeout_t=1;
    return;
  }
}


int ping_run(DWORD *address)
{
  *address=0;
  // wait for reply

  struct sockaddr_in from;
  int fromlen = sizeof (from);

  char packet[MAXPACKET];

  DWORD start_t=GetTickCount();
  if (m_pings_used > 0) for (;;)
  {
    int cc=recvfrom(s, packet, MAXPACKET, 0, (struct sockaddr *)&from, &fromlen);
    if (cc > 0)
    {
	    struct ip *ip;
	    long *lp = (long *) packet;
	    int hlen;
      int x;
      for (x = 0; x < m_pings_used; x ++)
      {
	      if (from.sin_addr.s_addr == m_pings[x].address)
        {
	        ip = (struct ip *) packet;
	        hlen = ip->ip_hl << 2;
	        if (cc >= hlen + ICMP_MINLEN) 
          {
          	struct icmp *icp = (struct icmp *) packet;
	          cc -= hlen;
	          icp = (struct icmp *)(packet + hlen);
            if (icp->icmp_type == ICMP_ECHOREPLY && icp->icmp_id == ident  && icp->icmp_seq == m_pings[x].seq)
            {
              DWORD t=m_pings[x].start_t;
              *address = m_pings[x].address;
              memcpy(m_pings+x,m_pings+x+1,sizeof(m_pings[0])*(--m_pings_used-x));
              return GetTickCount()-t;
            }
          }      
        }
      }
   //   if (GetTickCount()-start_t > 100) break; // dont run for more than 100ms
    }
    else break;
  }

  int x;
  for (x = 0; x < m_pings_used; x ++)
  {
    if (GetTickCount() > m_pings[x].start_t+m_pings[x].timeout_t)
    {
      *address = m_pings[x].address;
      memcpy(m_pings+x,m_pings+x+1,sizeof(m_pings[0])*(--m_pings_used-x));
      return -1;
    }
  }

  return -1;
}

void ping_quit()
{
  if (s)
    closesocket(s);
  s=0;
}
